import json

from six.moves import urllib

from dcos import config, cosmos, http, packagemanager, rpcclient, util
from dcos.errors import DCOSException

logger = util.get_logger(__name__)

EMBED_ACTIVE_RUNS = 'activeRuns'
EMBED_SCHEDULES = 'schedules'
EMBED_HISTORY = 'history'
EMBED_HISTORY_SUMMARY = 'historySummary'


def create_client(toml_config=None):
    """Creates a Metronome client with the supplied configuration.

    :param toml_config: configuration dictionary
    :type toml_config: config.Toml
    :returns: Metronome client
    :rtype: dcos.metronome.Client
    """

    if toml_config is None:
        toml_config = config.get_config()

    metronome_url = _get_metronome_url(toml_config)
    timeout = config.get_config_val('core.timeout') or http.DEFAULT_TIMEOUT
    rpc_client = rpcclient.create_client(metronome_url, timeout)

    logger.info('Creating metronome client with: %r', metronome_url)
    return Client(rpc_client)


def _get_embed_query_string(embed_list):
    return '?{}'.format('&'.join('embed=%s' % (item) for item in embed_list))


def _get_metronome_url(toml_config=None):
    """
    :param toml_config: configuration dictionary
    :type toml_config: config.Toml
    :returns: metronome base url
    :rtype: str
    """
    if toml_config is None:
        toml_config = config.get_config()

    metronome_url = config.get_config_val('metronome.url', toml_config)
    if metronome_url is None:
        # dcos must be capable to use dcos_url
        _check_capability()
        dcos_url = config.get_config_val('core.dcos_url', toml_config)
        if dcos_url is None:
            raise config.missing_config_exception(['core.dcos_url'])
        metronome_url = urllib.parse.urljoin(dcos_url, 'service/metronome/')

    return metronome_url


class Client(object):
    """Class for talking to the Metronome server.

    :param rpc_client: provides a method for making HTTP requests
    :type rpc_client: _RpcClient
    """

    def __init__(self, rpc_client):
        self._rpc = rpc_client

    def get_about(self):
        """Returns info about Metronome instance

        :returns Metronome information
        :rtype: dict
        """

        response = self._rpc.http_req(http.get, 'v1/info')

        return response.json()

    def get_job(self, job_id, embed_with=None):
        """Returns a representation of the requested job.

        :param job_id: the ID of the application
        :type job_id: str
        :param embed_with: list of strings to ?embed=str&embed=str2...
        :type embed_with: [str]
        :returns: the requested Metronome job
        :rtype: dict
        """
        # refactor util name it isn't marathon specific
        job_id = util.normalize_marathon_id_path(job_id)
        embeds = _get_embed_query_string(embed_with) if embed_with else None
        url = ('v1/jobs{}{}'.format(job_id, embeds)
               if embeds else 'v1/jobs{}'.format(job_id))
        response = self._rpc.http_req(http.get, url)
        return response.json()

    def get_jobs(self, embed_with=None):
        """Get a list of known jobs.

        :param embed_with: list of strings to ?embed=str&embed=str2...
        :type embed_with: [str]
        :returns: list of known jobs
        :rtype: [dict]
        """
        embeds = _get_embed_query_string(embed_with) if embed_with else None
        url = 'v1/jobs{}'.format(embeds) if embeds else 'v1/jobs'
        response = self._rpc.http_req(http.get, url)
        return response.json()

    def add_job(self, job_resource):
        """Add a new job.

        :param job_resource: job resource
        :type job_resource: dict, bytes or file
        :returns: the job description
        :rtype: dict
        """

        # The file type exists only in Python 2, preventing type(...) is file.
        if hasattr(job_resource, 'read'):
            job_json = json.load(job_resource)
        else:
            job_json = job_resource

        response = self._rpc.http_req(http.post, 'v1/jobs', json=job_json)
        # need deployment ID
        return response.json()

    def _update_req(
            self, resource_type, resource_id, resource_json, force=False):
        """Send an HTTP request to update an application, group, or pod.

        :param resource_type: one of 'apps', 'groups', or 'pods'
        :type resource_type: str
        :param resource_id: the app, group, or pod ID
        :type resource_id: str
        :param resource_json: the json payload
        :type resource_json: {}
        :param force: whether to override running deployments
        :type force: bool
        :returns: the response from Metronome
        :rtype: requests.Response
        """

        path_template = 'v1/{}/{{}}'.format(resource_type)
        path = self._job_id_path_format(path_template, resource_id)
        params = self._force_params(force)
        return self._rpc.http_req(
            http.put, path, params=params, json=resource_json)

    def _update(self, resource_type, resource_id, resource_json, force=False):
        """Update an application or group.

        The HTTP response is handled differently for pods; see `update_pod`.

        :param resource_type: either 'apps' or 'groups'
        :type resource_type: str
        :param resource_id: the app or group ID
        :type resource_id: str
        :param resource_json: the json payload
        :type resource_json: {}
        :param force: whether to override running deployments
        :type force: bool
        :returns: the resulting deployment ID
        :rtype: str
        """

        response = self._update_req(
            resource_type, resource_id, resource_json, force)
        body_json = self._parse_json(response)

        try:
            return body_json.get('deploymentId')
        except KeyError:
            template = ('Error: missing "deploymentId" field in the following '
                        'JSON response from Job:\n{}')
            rendered_json = json.dumps(body_json, indent=2, sort_keys=True)
            raise DCOSException(template.format(rendered_json))

    def update_job(self, job_id, payload, force=False):
        """Update a job.

        :param job_id: the job id
        :type job_id: str
        :param payload: the json payload
        :type payload: dict
        :param force: whether to override running deployments
        :type force: bool
        :returns: the resulting deployment ID
        :rtype: str
        """

        return self._update('jobs', job_id, payload, force)

    def remove_job(self, job_id, force=False):
        """Completely removes the requested application.

        :param job_id: the ID of the job to remove
        :type job_id: str
        :param force: whether to override running deployments
        :type force: bool
        :rtype: None
        """

        job_id = util.normalize_marathon_id_path(job_id)
        params = self._force_params(force)
        path = 'v1/jobs{}'.format(job_id)
        self._rpc.http_req(http.delete, path, params=params)

    def get_schedules(self, job_id):
        """Gets the schedules for a given job

        :param job_id: the ID of the job to remove
        :type job_id: str
        :rtype: json of schedules
        """
        job_id = util.normalize_marathon_id_path(job_id)
        path = 'v1/jobs{}/schedules'.format(job_id)
        response = self._rpc.http_req(http.get, path)
        return response.json()

    def get_schedule(self, job_id, schedule_id):
        """Gets the schedules for a given job

        :param job_id: the ID of the job to remove
        :type job_id: str
        :rtype: json of schedules
        """
        job_id = util.normalize_marathon_id_path(job_id)
        schedule_id = util.normalize_marathon_id_path(schedule_id)
        path = 'v1/jobs{}/schedules/{}'.format(job_id, schedule_id)
        response = self._rpc.http_req(http.get, path)
        return response.json()

    def add_schedule(self, job_id, schedule_resource):
        """Gets the schedules for a given job

        :param job_id: the ID of the job to remove
        :type job_id: str
        :rtype: json of schedules
        """
        job_id = util.normalize_marathon_id_path(job_id)
        if hasattr(schedule_resource, 'read'):
            schedule_json = json.load(schedule_resource)
        else:
            schedule_json = schedule_resource

        path = 'v1/jobs{}/schedules'.format(job_id)
        response = self._rpc.http_req(http.post, path, json=schedule_json)

        return response.json()

    def update_schedule(self, job_id, schedule_id, schedule_resource):
        """Gets the schedules for a given job

        :param job_id: the ID of the job to remove
        :type job_id: str
        :rtype: json of schedules
        """
        job_id = util.normalize_marathon_id_path(job_id)
        if hasattr(schedule_resource, 'read'):
            schedule_json = json.load(schedule_resource)
        else:
            schedule_json = schedule_resource

        path = 'v1/jobs{}/schedules/{}'.format(job_id, schedule_id)
        response = self._rpc.http_req(http.put, path, json=schedule_json)

        return response.json()

    def remove_schedule(self, job_id, schedule_id):
        """Completely removes the requested application.

        :param job_id: the ID of the job to remove
        :type job_id: str
        :param force: whether to override running deployments
        :type force: bool
        :rtype: None
        """

        job_id = util.normalize_marathon_id_path(job_id)
        schedule_id = util.normalize_marathon_id_path(schedule_id)
        path = 'v1/jobs{}/schedules/{}'.format(job_id, schedule_id)
        self._rpc.http_req(http.delete, path)

    def run_job(self, job_id):
        """Add a new job.

        :param job_id: the ID of the job to remove
        :type job_id: str
        :rtype: None
        """

        job_id = util.normalize_marathon_id_path(job_id)
        path = '/v1/jobs{}/runs'.format(job_id)
        response = self._rpc.http_req(http.post, path)

        return response.json()

    def get_runs(self, job_id):
        """Gets the schedules for a given job

        :param job_id: the ID of the job to remove
        :type job_id: str
        :rtype: json of schedules
        """
        job_id = util.normalize_marathon_id_path(job_id)
        path = '/v1/jobs{}/runs'.format(job_id)
        response = self._rpc.http_req(http.get, path)
        return response.json()

    def get_run(self, job_id, run_id):
        """Add a new job.

        :param job_id: the ID of the job
        :type job_id: str
        :param run_id: the ID of the job run
        :type run_id: str
        :rtype: None
        """

        job_id = util.normalize_marathon_id_path(job_id)
        run_id = util.normalize_marathon_id_path(run_id)
        path = '/v1/jobs{}/runs{}'.format(job_id, run_id)
        response = self._rpc.http_req(http.get, path)
        return response.json()

    def kill_run(self, job_id, run_id):
        """Add a new job.

        :param job_id: the ID of the job
        :type job_id: str
        :param run_id: the ID of the job run to remove
        :type run_id: str
        :rtype: None
        """

        job_id = util.normalize_marathon_id_path(job_id)
        run_id = util.normalize_marathon_id_path(run_id)
        path = '/v1/jobs{}/runs{}/actions/stop'.format(job_id, run_id)
        self._rpc.http_req(http.post, path)

    @staticmethod
    def _job_id_path_format(url_path_template, id_path):
        """Substitutes a Metronome "ID path" into a URL path format string,
        ensuring the result is well-formed.

        All leading and trailing slashes in the ID will be removed, and the ID
        will have all URL-unsafe characters escaped, as if by
        urllib.parse.quote().

        :param url_path_template: format string for the path portion of a URL,
                                  with a single format specifier (i.e. {})
                                  where the "ID path" should be inserted
        :type url_path_template: str
        :param id_path: a Job "ID path", e.g. app, group, or pod ID
        :type id_path: str
        :returns: the url path template with the ID inserted
        :rtype: str
        """

        normalized_id_path = urllib.parse.quote(id_path.strip('/'))
        return url_path_template.format(normalized_id_path)

    @staticmethod
    def _force_params(force):
        """Returns the query parameters that signify the provided force value.

        :param force: whether to override running deployments
        :type force: bool
        :rtype: {} | None
        """

        return {'stopCurrentJobRuns': 'true'} if force else None

    @staticmethod
    def _parse_json(response):
        """Attempts to parse the body of the given response as JSON.

        Raises DCOSException if parsing fails.

        :param response: the response containing the body to parse
        :type response: requests.Response
        :return: the parsed JSON
        :rtype: {} | [] | str | int | float | bool | None
        """

        try:
            return response.json()
        except:
            template = ('Error: Response from Metronome was not in expected '
                        'JSON format:\n{}')
            raise DCOSException(template.format(response.text))


def _check_capability():
    """
    The function checks if cluster has metronome capability.

    :raises: DCOSException if cluster does not have metronome capability
    """

    manager = packagemanager.PackageManager(cosmos.get_cosmos_url())
    if not manager.has_capability('METRONOME'):
        raise DCOSException(
            'DC/OS backend does not support metronome capabilities in this '
            'version. Must be DC/OS >= 1.8')



import json
import re

from .helpers.common import assert_command, assert_lines, exec_command
from .helpers.marathon import (group, remove_group, show_app,
                               watch_all_deployments)


GOOD_GROUP = 'tests/data/marathon/groups/good.json'
SCALE_GROUP = 'tests/data/marathon/groups/scale.json'


def test_add_group_by_stdin():
    _add_group_by_stdin(GOOD_GROUP)
    remove_group('test-group')


def test_group_list_table():
    with group(GOOD_GROUP, 'test-group'):
        assert_lines(['dcos', 'marathon', 'group', 'list'], 3)


def test_validate_complicated_group_and_app():
    with group('tests/data/marathon/groups/complicated.json', 'test-group'):
        pass


def test_add_existing_group():
    with group(GOOD_GROUP, 'test-group'):
        with open(GOOD_GROUP) as fd:
            stderr = b"Group '/test-group' already exists\n"
            assert_command(['dcos', 'marathon', 'group', 'add'],
                           returncode=1,
                           stderr=stderr,
                           stdin=fd)


def test_add_bad_complicated_group():
    with open('tests/data/marathon/groups/complicated_bad.json') as fd:
        returncode, stdout, stderr = exec_command(
            ['dcos', 'marathon', 'group', 'add'],
            stdin=fd)

        stderr_end = b"""{"message":"Invalid JSON","details":[{"path":"/groups(0)/apps(0)/id","errors":["\'id\' is undefined on object: {}"]}]}"""  # noqa: E501

        assert returncode == 1
        assert stderr_end in stderr
        assert stdout == b''


def test_update_group_from_stdin():
    with group(GOOD_GROUP, 'test-group'):
        _update_group(
            'test-group',
            'tests/data/marathon/groups/update_good.json')
        show_app('test-group/updated')


def test_update_missing_group():
    assert_command(['dcos', 'marathon', 'group', 'update', 'missing-id'],
                   stderr=b"Error: Group '/missing-id' does not exist\n",
                   returncode=1)


def test_scale_group():
    with group(SCALE_GROUP, 'scale-group'):
        returncode, stdout, stderr = exec_command(
            ['dcos', 'marathon', 'group', 'scale', 'scale-group', '2'])
        assert stderr == b''
        assert returncode == 0
        watch_all_deployments()
        returncode, stdout, stderr = exec_command(
            ['dcos', 'marathon', 'group', 'show', 'scale-group'])
        res = json.loads(stdout.decode('utf-8'))

        assert res['groups'][0]['apps'][0]['instances'] == 2


def test_scale_group_not_exist():
    returncode, stdout, stderr = exec_command(['dcos', 'marathon', 'group',
                                               'scale', 'scale-group', '2'])
    assert stderr == b''
    watch_all_deployments()
    returncode, stdout, stderr = exec_command(
        ['dcos', 'marathon', 'group', 'show',
         'scale-group'])
    res = json.loads(stdout.decode('utf-8'))

    assert len(res['apps']) == 0
    remove_group('scale-group')


def test_scale_group_when_scale_factor_negative():
    with group(SCALE_GROUP, 'scale-group'):
        returncode, stdout, stderr = exec_command(
            ['dcos', 'marathon', 'group', 'scale', 'scale-group', '-2'])
        assert b'Command not recognized' in stdout
        assert returncode == 1


def test_scale_group_when_scale_factor_not_float():
    with group(SCALE_GROUP, 'scale-group'):
        returncode, stdout, stderr = exec_command(
            ['dcos', 'marathon', 'group', 'scale', 'scale-group', '1.a'])
        assert stderr == b'Error parsing string as float\n'
        assert returncode == 1


def _add_group_by_stdin(file_path):
    with open(file_path) as fd:
        cmd = ['dcos', 'marathon', 'group', 'add']
        returncode, stdout, stderr = exec_command(cmd, stdin=fd)
        assert returncode == 0
        assert re.fullmatch('Created deployment \S+\n',
                            stdout.decode('utf-8'))
        assert stderr == b''

    # Let's make sure that we don't return until the deployment has finished
    watch_all_deployments()


def _update_group(group_id, file_path):
    with open(file_path) as fd:
        returncode, stdout, stderr = exec_command(
            ['dcos', 'marathon', 'group', 'update', group_id],
            stdin=fd)

        assert returncode == 0
        assert stdout.decode().startswith('Created deployment ')
        assert stderr == b''

    # Let's make sure that we don't return until the deployment has finished
    watch_all_deployments()



import contextlib
import json
import os
import re
import shutil
import tempfile
import time

import dcoscli
from dcos import util
from .helpers.common import (assert_command, exec_command, file_json_ast,
                             zip_contents_as_json)
from .helpers.marathon import watch_all_deployments

command_base = ['dcos', 'experimental']
data_dir = os.path.join(os.getcwd(), 'tests', 'data')
build_data_dir = os.path.join(data_dir, 'package_build')


def runnable_package_path(index):
    return os.path.join(
        build_data_dir, 'helloworld', 'helloworld{}.json'.format(index))


def test_experimental():
    command = command_base + ['--help']
    with open('dcoscli/data/help/experimental.txt') as content:
        assert_command(command, stdout=content.read().encode())


def test_info():
    command = command_base + ['--info']
    out = b'Manage commands that are under development\n'
    assert_command(command, stdout=out)


def test_version():
    command = command_base + ['--version']
    out = b'dcos-experimental version SNAPSHOT\n'
    assert_command(command, stdout=out)


def test_package_build_with_only_resources():
    _successful_package_build_test(
        os.path.join(
            build_data_dir,
            "package_resource_only_reference.json"),
        expected_package_path=os.path.join(
            build_data_dir,
            "package_resource_only_reference_expected.json"))


def test_package_build_with_only_config_with_no_references():
    _successful_package_build_test(
        os.path.join(
            build_data_dir,
            "package_config_reference_expected.json"),
        expected_package_path=os.path.join(
            build_data_dir,
            "package_config_reference_expected.json"))


def test_package_build_with_only_config():
    _successful_package_build_test(
        os.path.join(
            build_data_dir,
            "package_config_reference.json"),
        expected_package_path=os.path.join(
            build_data_dir,
            "package_config_reference_expected.json"))


def test_package_build_with_only_marathon():
    _successful_package_build_test(
        os.path.join(
            build_data_dir,
            "package_marathon_reference.json"),
        expected_package_path=os.path.join(
            build_data_dir,
            "package_marathon_reference_expected.json"))


def test_package_build_with_only_resources_reference():
    _successful_package_build_test(
        os.path.join(
            build_data_dir,
            "package_resource_reference.json"))


def test_package_build_with_no_references():
    _successful_package_build_test(
        os.path.join(
            build_data_dir,
            "package_no_references.json"))


def test_package_build_with_all_references():
    _successful_package_build_test(
        os.path.join(
            build_data_dir,
            "package_all_references.json"))


def test_package_build_with_all_references_json():
    _successful_package_build_test(
        os.path.join(
            build_data_dir,
            "package_all_references.json"),
        expects_json=True)


def test_package_build_where_build_definition_does_not_exist():
    with _temporary_directory() as output_directory:
        build_definition_path = os.path.join(build_data_dir,
                                             "does_not_exist.json")
        stderr = ("The file [{}] does not exist\n"
                  .format(build_definition_path)
                  .encode())
        _package_build_failure(build_definition_path,
                               output_directory,
                               stderr=stderr)


def test_package_build_where_project_is_missing_references():
    with _temporary_directory() as output_directory:
        build_definition_path = (
            os.path.join(build_data_dir,
                         "package_missing_references.json"))
        marathon_json_path = os.path.join(build_data_dir,
                                          "marathon.json")
        stderr = ("Error opening file [{}]: No such file or directory\n"
                  .format(marathon_json_path)
                  .encode())
        _package_build_failure(build_definition_path,
                               output_directory,
                               stderr=stderr)


def test_package_build_where_reference_does_not_match_schema():
    with _temporary_directory() as output_directory:
        build_definition_path = os.path.join(
            build_data_dir,
            "package_reference_does_not_match_schema.json"
        )
        bad_resource_path = os.path.join(
            build_data_dir,
            "resource-bad.json"
        )
        stderr = ("Error validating package: "
                  "[{}] does not conform to the specified schema\n"
                  .format(bad_resource_path)
                  .encode())
        _package_build_failure(build_definition_path,
                               output_directory,
                               stderr=stderr)


def test_package_build_where_build_definition_does_not_match_schema():
    with _temporary_directory() as output_directory:
        bad_build_definition_path = os.path.join(
            build_data_dir,
            "package_no_match_schema.json"
        )
        stderr = ("Error validating package: "
                  "[{}] does not conform to the specified schema\n"
                  .format(bad_build_definition_path)
                  .encode())
        _package_build_failure(bad_build_definition_path,
                               output_directory,
                               stderr=stderr)


def test_package_build_where_build_definition_has_badly_formed_reference():
    with _temporary_directory() as output_directory:
        bad_build_definition_path = os.path.join(
            build_data_dir,
            "package_badly_formed_reference.json"
        )
        stderr = ("Error validating package: "
                  "[{}] does not conform to the specified schema\n"
                  .format(bad_build_definition_path)
                  .encode())
        _package_build_failure(bad_build_definition_path,
                               output_directory,
                               stderr=stderr)


def test_package_add_argument_exclussion():
    command = command_base + ['package', 'add',
                              '--dcos-package', runnable_package_path(1),
                              '--package-version', '3.0']
    code, out, err = exec_command(command)
    assert code == 1
    assert err == b''

    stdout = out.decode()
    not_recognized = 'Command not recognized'
    assert not_recognized in stdout


def test_service_start_happy_path():
    with _temporary_directory() as output_directory:
        runnable_package = _package_build(
            runnable_package_path(2), output_directory)
        name, version = _package_add(runnable_package)
        try:
            _service_start(name, version)
        finally:
            _service_stop(name)


def test_service_start_happy_path_json():
    with _temporary_directory() as output_directory:
        runnable_package = _package_build(
            runnable_package_path(3), output_directory)
        name, version = _package_add(runnable_package, expects_json=True)
        try:
            _service_start(name, version, expects_json=True)
        finally:
            _service_stop(name)


def test_service_start_happy_path_from_universe():
    package_name = 'hello-world'
    name, version = _package_add_universe(package_name)
    try:
        _service_start(name, version)
    finally:
        _service_stop(name)


def test_service_start_happy_path_from_universe_json():
    package_name = 'cassandra'
    name, version = _package_add_universe(package_name, expects_json=True)
    try:
        _service_start(name, version)
    finally:
        _service_stop(name)


def test_service_start_by_starting_same_service_twice():
    name, version = _package_add_universe('kafka')
    try:
        _service_start(name, version)
        stderr = b'The DC/OS service has already been started\n'
        _service_start_failure(name, version, stderr=stderr)
    finally:
        _service_stop(name)


def test_service_start_by_starting_service_not_added():
    stderr = b'Package [foo] not found\n'
    _service_start_failure('foo', stderr=stderr)


def _service_stop_cmd(package_name):
    return ['dcos', 'package', 'uninstall', package_name, '--yes']


def _service_list_cmd():
    return ['dcos', 'package', 'list', '--json']


def _service_start_cmd(package_name,
                       package_version=None,
                       options=None,
                       json=False):
    return (command_base +
            (['service', 'start']) +
            (['--json'] if json else []) +
            ([package_name]) +
            (['--package-version', package_version]
             if package_version else []) +
            (['--options', options] if options else []))


def _package_add_cmd(dcos_package=None,
                     package_name=None,
                     package_version=None,
                     json=False):
    return (command_base +
            (['package', 'add']) +
            (['--json'] if json else []) +
            (['--dcos-package', dcos_package] if dcos_package else []) +
            (['--package-name', package_name] if package_name else []) +
            (['--package-version', package_version]
             if package_version else []))


def _package_build_cmd(build_definition,
                       output_directory=None,
                       expects_json=False):
    return (command_base +
            (['package', 'build']) +
            (['--json'] if expects_json else []) +
            (['--output-directory', output_directory]
             if output_directory else []) +
            ([build_definition]))


def _service_stop(package_name):
    command = _service_stop_cmd(package_name)
    exec_command(command)
    watch_all_deployments()


def _service_list():
    command = _service_list_cmd()
    code, out, err = exec_command(command)
    assert code == 0
    assert err == b''
    return json.loads(out.decode())


def _service_start(package_name,
                   package_version,
                   options=None,
                   expects_json=False):
    command = _service_start_cmd(package_name,
                                 package_version,
                                 options,
                                 json=expects_json)
    code = 1
    max_retries = 10
    retry_number = 0
    while code != 0:
        code, out, err = exec_command(command)
        assert retry_number != max_retries, \
            'Waiting for package add to complete took too long'
        retry_number += 1
        time.sleep(5)

    assert code == 0
    assert err == b''

    if expects_json:
        expected = {
            'packageName': package_name,
            'packageVersion': package_version
        }
        actual = json.loads(out.decode())
        actual.pop('appId')
        assert expected == actual, (expected, actual)
    else:
        stdout = 'The service [{}] version [{}] has been started\n'.format(
            package_name, package_version).encode()
        assert out == stdout, (out, stdout)

    running_services = _service_list()
    assert package_name in map(lambda pkg: pkg['name'], running_services)


def _service_start_failure(package_name,
                           package_version=None,
                           options=None,
                           return_code=1,
                           stdout=b'',
                           stderr=b''):
    command = _service_start_cmd(package_name,
                                 package_version,
                                 options)
    assert_command(command,
                   returncode=return_code,
                   stdout=stdout,
                   stderr=stderr)


def _package_add(package, expects_json=False):
    command = _package_add_cmd(dcos_package=package, json=expects_json)
    code, out, err = exec_command(command)
    assert code == 0
    assert err == b''
    if expects_json:
        metadata = json.loads(out.decode())
        metadata.pop('releaseVersion')
        assert metadata == zip_contents_as_json(package, 'metadata.json')
    else:
        metadata = zip_contents_as_json(package, 'metadata.json')
        stdout = (
            'The package [{}] version [{}] has been added to DC/OS\n'.format(
                metadata['name'], metadata['version'])).encode()
        assert out == stdout, (out, stdout)

    return metadata['name'], metadata['version']


def _package_add_universe(package_name,
                          package_version=None,
                          expects_json=False):
    command = _package_add_cmd(package_name=package_name,
                               package_version=package_version,
                               json=expects_json)
    code, out, err = exec_command(command)
    assert code == 0
    assert err == b''
    if expects_json:
        metadata = json.loads(out.decode())
        name = metadata['name']
        version = metadata['version']
    else:
        name_version = re.search("\[(.*)\].*\[(.*)\]", out.decode())
        name = name_version.group(1)
        version = name_version.group(2)
        stdout = (
            'The package [{}] version [{}] has been added to DC/OS\n'.format(
                name, version)).encode()
        assert out == stdout, (out, stdout)

    assert name == package_name
    assert version == package_version if package_version else True

    return name, version


def _package_build(build_definition_path,
                   output_directory,
                   metadata=None,
                   manifest=None,
                   expects_json=False):
    command = _package_build_cmd(build_definition_path,
                                 output_directory,
                                 expects_json=expects_json)

    code, out, err = exec_command(command)
    assert code == 0
    assert err == b''

    out_str = out.decode()
    if expects_json:
        out_json = json.loads(out_str)
        assert out_json, out_str
        package_path = out_json.get('package_path')
    else:
        assert out_str.startswith("Created DC/OS Universe Package")
        package_path = re.search("\[(.*)\]", out_str).group(1)

    assert package_path, out_str
    assert os.path.exists(package_path)

    name, version, md5 = _decompose_name(package_path)
    build_definition = file_json_ast(build_definition_path)
    assert name == build_definition['name']
    assert version == build_definition['version']
    assert md5 == _get_md5_hash(package_path)
    assert (manifest is None or
            manifest == zip_contents_as_json(package_path, 'manifest.json'))
    assert (metadata is None or
            metadata == zip_contents_as_json(package_path, 'metadata.json'))

    return package_path


def _package_build_failure(build_definition_path,
                           output_directory,
                           return_code=1,
                           stdout=b'',
                           stderr=b''):
    command = _package_build_cmd(build_definition_path, output_directory)

    assert_command(command,
                   returncode=return_code,
                   stdout=stdout,
                   stderr=stderr)
    assert len(os.listdir(output_directory)) == 0


def _successful_package_build_test(
        build_definition_path,
        expected_package_path=os.path.join(
            build_data_dir,
            "package_no_references.json"),
        expects_json=False):
    with _temporary_directory() as output_directory:
        metadata = file_json_ast(expected_package_path)
        manifest = {
            'built-by': "dcoscli.version={}".format(dcoscli.version)
        }
        _package_build(build_definition_path,
                       output_directory,
                       metadata=metadata,
                       manifest=manifest,
                       expects_json=expects_json)


def _decompose_name(package_path):
    parts = re.search(
        '^([^-]+)-(.+)-([^-]+)\.dcos',
        os.path.basename(package_path))
    assert parts is not None, package_path
    return parts.group(1), parts.group(2), parts.group(3)


def _get_md5_hash(path):
    with open(path, 'rb') as f:
        return util.md5_hash_file(f)


@contextlib.contextmanager
def _temporary_directory():
    tmp_dir = tempfile.mkdtemp()
    try:
        yield tmp_dir
    finally:
        shutil.rmtree(tmp_dir)



import json
import os
import sys
import time

import docopt
import pkg_resources
import six

import dcoscli
from dcos import cmds, emitting, http, jsonitem, marathon, options, util
from dcos.errors import DCOSException
from dcoscli import tables
from dcoscli.subcommand import default_command_info, default_doc
from dcoscli.util import decorate_docopt_usage

logger = util.get_logger(__name__)
emitter = emitting.FlatEmitter()


def main(argv):
    try:
        return _main(argv)
    except DCOSException as e:
        emitter.publish(e)
        return 1


@decorate_docopt_usage
def _main(argv):
    args = docopt.docopt(
        default_doc("marathon"),
        argv=argv,
        version='dcos-marathon version {}'.format(dcoscli.version))

    return cmds.execute(_cmds(), args)


def _cmds():
    """
    :returns: all the supported commands
    :rtype: dcos.cmds.Command
    """

    subcommand = MarathonSubcommand()

    return [
        cmds.Command(
            hierarchy=['marathon', 'version', 'list'],
            arg_keys=['<app-id>', '--max-count'],
            function=subcommand.version_list),

        cmds.Command(
            hierarchy=['marathon', 'deployment', 'list'],
            arg_keys=['<app-id>', '--json'],
            function=subcommand.deployment_list),

        cmds.Command(
            hierarchy=['marathon', 'deployment', 'rollback'],
            arg_keys=['<deployment-id>'],
            function=subcommand.deployment_rollback),

        cmds.Command(
            hierarchy=['marathon', 'deployment', 'stop'],
            arg_keys=['<deployment-id>'],
            function=subcommand.deployment_stop),

        cmds.Command(
            hierarchy=['marathon', 'deployment', 'watch'],
            arg_keys=['<deployment-id>', '--max-count', '--interval'],
            function=subcommand.deployment_watch),

        cmds.Command(
            hierarchy=['marathon', 'task', 'list'],
            arg_keys=['<app-id>', '--json'],
            function=subcommand.task_list),

        cmds.Command(
            hierarchy=['marathon', 'task', 'stop'],
            arg_keys=['<task-id>', '--wipe'],
            function=subcommand.task_stop),

        cmds.Command(
            hierarchy=['marathon', 'task', 'kill'],
            arg_keys=['<task-ids>', '--scale', '--wipe', '--json'],
            function=subcommand.task_kill),

        cmds.Command(
            hierarchy=['marathon', 'task', 'show'],
            arg_keys=['<task-id>'],
            function=subcommand.task_show),

        cmds.Command(
            hierarchy=['marathon', 'app', 'add'],
            arg_keys=['<app-resource>'],
            function=subcommand.add),

        cmds.Command(
            hierarchy=['marathon', 'app', 'list'],
            arg_keys=['--json'],
            function=subcommand.list),

        cmds.Command(
            hierarchy=['marathon', 'app', 'remove'],
            arg_keys=['<app-id>', '--force'],
            function=subcommand.remove),

        cmds.Command(
            hierarchy=['marathon', 'app', 'show'],
            arg_keys=['<app-id>', '--app-version'],
            function=subcommand.show),

        cmds.Command(
            hierarchy=['marathon', 'app', 'start'],
            arg_keys=['<app-id>', '<instances>', '--force'],
            function=subcommand.start),

        cmds.Command(
            hierarchy=['marathon', 'app', 'stop'],
            arg_keys=['<app-id>', '--force'],
            function=subcommand.stop),

        cmds.Command(
            hierarchy=['marathon', 'app', 'update'],
            arg_keys=['<app-id>', '<properties>', '--force'],
            function=subcommand.update),

        cmds.Command(
            hierarchy=['marathon', 'app', 'restart'],
            arg_keys=['<app-id>', '--force'],
            function=subcommand.restart),

        cmds.Command(
            hierarchy=['marathon', 'app', 'kill'],
            arg_keys=['<app-id>', '--scale', '--host'],
            function=subcommand.kill),

        cmds.Command(
            hierarchy=['marathon', 'group', 'add'],
            arg_keys=['<group-resource>'],
            function=subcommand.group_add),

        cmds.Command(
            hierarchy=['marathon', 'group', 'list'],
            arg_keys=['--json'],
            function=subcommand.group_list),

        cmds.Command(
            hierarchy=['marathon', 'group', 'show'],
            arg_keys=['<group-id>', '--group-version'],
            function=subcommand.group_show),

        cmds.Command(
            hierarchy=['marathon', 'group', 'remove'],
            arg_keys=['<group-id>', '--force'],
            function=subcommand.group_remove),

        cmds.Command(
            hierarchy=['marathon', 'group', 'update'],
            arg_keys=['<group-id>', '<properties>', '--force'],
            function=subcommand.group_update),

        cmds.Command(
            hierarchy=['marathon', 'group', 'scale'],
            arg_keys=['<group-id>', '<scale-factor>', '--force'],
            function=subcommand.group_scale),

        cmds.Command(
            hierarchy=['marathon', 'pod', 'add'],
            arg_keys=['<pod-resource>'],
            function=subcommand.pod_add),

        cmds.Command(
            hierarchy=['marathon', 'pod', 'remove'],
            arg_keys=['<pod-id>', '--force'],
            function=subcommand.pod_remove),

        cmds.Command(
            hierarchy=['marathon', 'pod', 'list'],
            arg_keys=['--json'],
            function=subcommand.pod_list),

        cmds.Command(
            hierarchy=['marathon', 'pod', 'show'],
            arg_keys=['<pod-id>'],
            function=subcommand.pod_show),

        cmds.Command(
            hierarchy=['marathon', 'pod', 'update'],
            arg_keys=['<pod-id>', '--force'],
            function=subcommand.pod_update),

        cmds.Command(
            hierarchy=['marathon', 'pod', 'kill'],
            arg_keys=['<pod-id>', '<instance-ids>'],
            function=subcommand.pod_kill),

        cmds.Command(
            hierarchy=['marathon', 'debug', 'list'],
            arg_keys=['--json'],
            function=subcommand.debug_list),

        cmds.Command(
            hierarchy=['marathon', 'debug', 'summary'],
            arg_keys=['<app-id>', '--json'],
            function=subcommand.debug_summary),

        cmds.Command(
            hierarchy=['marathon', 'debug', 'details'],
            arg_keys=['<app-id>', '--json'],
            function=subcommand.debug_details),

        cmds.Command(
            hierarchy=['marathon', 'about'],
            arg_keys=[],
            function=subcommand.about),

        cmds.Command(
            hierarchy=['marathon'],
            arg_keys=['--config-schema', '--info'],
            function=_marathon)
    ]


def _marathon(config_schema, info):
    """
    :param config_schema: Whether to output the config schema
    :type config_schema: boolean
    :param info: Whether to output a description of this subcommand
    :type info: boolean
    :returns: process return code
    :rtype: int
    """

    if config_schema:
        schema = _cli_config_schema()
        emitter.publish(schema)
    elif info:
        _info()
    else:
        doc = default_command_info("marathon")
        emitter.publish(options.make_generic_usage_message(doc))
        return 1

    return 0


def _info():
    """
    :returns: process return code
    :rtype: int
    """

    emitter.publish(default_command_info("marathon"))
    return 0


class ResourceReader(object):
    """Encapsulates side-effecting methods for loading Marathon resources."""

    @staticmethod
    def get_resource(name):
        """
        :param name: optional filename or http(s) url
        for the application or group resource
        :type name: str | None
        :returns: resource
        :rtype: dict
        """
        if name is not None:
            if os.path.isfile(name):
                with util.open_file(name) as resource_file:
                    return util.load_json(resource_file)
            else:
                try:
                    http.silence_requests_warnings()
                    req = http.get(name)
                    if req.status_code == 200:
                        data = b''
                        for chunk in req.iter_content(1024):
                            data += chunk
                        return util.load_jsons(data.decode('utf-8'))
                    else:
                        raise Exception
                except Exception:
                    logger.exception('Cannot read from resource %s', name)
                    raise DCOSException(
                        "Can't read from resource: {0}.\n"
                        "Please check that it exists.".format(name))

        example = "E.g.: dcos marathon app add < app_resource.json"
        ResourceReader._assert_no_tty(example)

        return util.load_json(sys.stdin)

    @staticmethod
    def get_resource_from_properties(properties):
        """
        :param properties: JSON items in the form key=value
        :type properties: [str]
        :returns: resource JSON
        :rtype: dict
        """

        if len(properties) == 0:
            example =\
                "E.g. dcos marathon app update your-app-id < app_update.json"
            ResourceReader._assert_no_tty(example)

            return util.load_jsons(sys.stdin.read())

        resource_json = {}
        for prop in properties:
            key, value = jsonitem.parse_json_item(prop, None)

            key = jsonitem.clean_value(key)
            if key in resource_json:
                raise DCOSException(
                    'Key {!r} was specified more than once'.format(key))

            resource_json[key] = value
        return resource_json

    @staticmethod
    def _assert_no_tty(command_example):
        if sys.stdin.isatty():
            # We don't support TTY right now.
            # In the future we will start an editor
            template = ("We currently don't support reading from the TTY. "
                        "Please specify an application JSON.\n{}")
            raise DCOSException(template.format(command_example))


class MarathonSubcommand(object):
    """Defines a method for each operation of the `dcos marathon` subcommand.

    :param resource_reader: provides input methods for Marathon resources
    :type resource_reader: ResourceReader
    :param create_marathon_client: a callable that returns an instance of
                                   marathon.Client
    :type create_marathon_client: collections.abc.Callable
    """

    def __init__(self,
                 resource_reader=ResourceReader(),
                 create_marathon_client=marathon.create_client):
        self._resource_reader = resource_reader
        self._create_marathon_client = create_marathon_client

    def about(self):
        """
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()

        emitter.publish(client.get_about())
        return 0

    def add(self, app_resource):
        """
        :param app_resource: optional filename for the application resource
        :type app_resource: str
        :returns: process return code
        :rtype: int
        """
        application_resource = self._resource_reader.get_resource(app_resource)

        # Add application to marathon
        client = self._create_marathon_client()

        # Check that the application doesn't exist
        app_id = util.normalize_marathon_id_path(application_resource['id'])

        try:
            client.get_app(app_id)
        except DCOSException as e:
            logger.exception(e)
        else:
            message = "Application '{}' already exists".format(app_id)
            raise DCOSException(message)

        deployment = client.add_app(application_resource)
        emitter.publish('Created deployment {}'.format(deployment))

        return 0

    def list(self, json_):
        """
        :param json_: output json if True
        :type json_: bool
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()
        apps = client.get_apps()

        if json_:
            emitter.publish(apps)
        else:
            deployments = client.get_deployments()
            queued_apps = client.get_queued_apps()
            _enhance_row_with_overdue_information(apps, queued_apps)
            table = tables.app_table(apps, deployments)
            output = six.text_type(table)
            if output:
                emitter.publish(output)

        return 0

    def group_list(self, json_):
        """
        :param json_: output json if True
        :type json_: bool
        :rtype: int
        :returns: process return code
        """

        client = self._create_marathon_client()
        groups = client.get_groups()

        emitting.publish_table(emitter, groups, tables.group_table, json_)
        return 0

    def group_add(self, group_resource):
        """
        :param group_resource: optional filename for the group resource
        :type group_resource: str
        :returns: process return code
        :rtype: int
        """

        group_resource = self._resource_reader.get_resource(group_resource)

        client = self._create_marathon_client()

        # Check that the group doesn't exist
        group_id = util.normalize_marathon_id_path(group_resource['id'])

        try:
            client.get_group(group_id)
        except DCOSException as e:
            logger.exception(e)
        else:
            raise DCOSException("Group '{}' already exists".format(group_id))

        deployment = client.create_group(group_resource)
        emitter.publish('Created deployment {}'.format(deployment))

        return 0

    def remove(self, app_id, force):
        """
        :param app_id: ID of the app to remove
        :type app_id: str
        :param force: Whether to override running deployments.
        :type force: bool
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()
        client.remove_app(app_id, force)
        return 0

    def group_remove(self, group_id, force):
        """
        :param group_id: ID of the app to remove
        :type group_id: str
        :param force: Whether to override running deployments.
        :type force: bool
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()
        client.remove_group(group_id, force)
        return 0

    def show(self, app_id, version):
        """Show details of a Marathon application.

        :param app_id: The id for the application
        :type app_id: str
        :param version: The version, either absolute (date-time) or relative
        :type version: str
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()

        if version is not None:
            version = _calculate_version(client, app_id, version)

        app = client.get_app(app_id, version=version)

        emitter.publish(app)
        return 0

    def group_show(self, group_id, version=None):
        """Show details of a Marathon application.

        :param group_id: The id for the application
        :type group_id: str
        :param version: The version, either absolute (date-time) or relative
        :type version: str
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()

        app = client.get_group(group_id, version=version)

        emitter.publish(app)
        return 0

    def group_update(self, group_id, properties, force):
        """
        :param group_id: the id of the group
        :type group_id: str
        :param properties: json items used to update group
        :type properties: [str]
        :param force: whether to override running deployments
        :type force: bool
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()

        # Ensure that the group exists
        client.get_group(group_id)

        resource = self._resource_reader.\
            get_resource_from_properties(properties)
        deployment = client.update_group(group_id, resource, force)

        emitter.publish('Created deployment {}'.format(deployment))
        return 0

    def start(self, app_id, instances, force):
        """Start a Marathon application.

        :param app_id: the id for the application
        :type app_id: str
        :param instances: the number of instances to start
        :type instances: str
        :param force: whether to override running deployments
        :type force: bool
        :returns: process return code
        :rtype: int
        """

        # Check that the application exists
        client = self._create_marathon_client()

        desc = client.get_app(app_id)

        if desc['instances'] > 0:
            emitter.publish(
                'Application {!r} already started: {!r} instances.'.format(
                    app_id,
                    desc['instances']))
            return 1

        # Need to add the 'id' because it is required
        app_json = {'id': app_id}

        # Set instances to 1 if not specified
        if instances is None:
            instances = 1
        else:
            instances = util.parse_int(instances)
            if instances <= 0:
                emitter.publish(
                    'The number of instances must be positive: {!r}.'.format(
                        instances))
                return 1

        app_json['instances'] = instances

        deployment = client.update_app(app_id, app_json, force)

        emitter.publish('Created deployment {}'.format(deployment))

        return 0

    def stop(self, app_id, force):
        """Stop a Marathon application

        :param app_id: the id of the application
        :type app_id: str
        :param force: whether to override running deployments
        :type force: bool
        :returns: process return code
        :rtype: int
        """

        # Check that the application exists
        client = self._create_marathon_client()

        desc = client.get_app(app_id)

        if desc['instances'] <= 0:
            emitter.publish(
                'Application {!r} already stopped: {!r} instances.'.format(
                    app_id,
                    desc['instances']))
            return 1

        app_json = {'instances': 0}

        deployment = client.update_app(app_id, app_json, force)

        emitter.publish('Created deployment {}'.format(deployment))

    def update(self, app_id, properties, force):
        """
        :param app_id: the id of the application
        :type app_id: str
        :param properties: json items used to update resource
        :type properties: [str]
        :param force: whether to override running deployments
        :type force: bool
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()

        # Ensure that the application exists
        client.get_app(app_id)

        resource = self._resource_reader.\
            get_resource_from_properties(properties)
        deployment = client.update_app(app_id, resource, force)

        emitter.publish('Created deployment {}'.format(deployment))
        return 0

    def group_scale(self, group_id, scale_factor, force):
        """
        :param group_id: the id of the group
        :type group_id: str
        :param scale_factor: scale factor for application group
        :type scale_factor: str
        :param force: whether to override running deployments
        :type force: bool
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()
        scale_factor = util.parse_float(scale_factor)
        deployment = client.scale_group(group_id, scale_factor, force)
        emitter.publish('Created deployment {}'.format(deployment))
        return 0

    def restart(self, app_id, force):
        """
        :param app_id: the id of the application
        :type app_id: str
        :param force: whether to override running deployments
        :type force: bool
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()

        desc = client.get_app(app_id)

        if desc['instances'] <= 0:
            app_id = util.normalize_marathon_id_path(app_id)
            emitter.publish(
                'Unable to perform rolling restart of application {!r} '
                'because it has no running tasks'.format(
                    app_id,
                    desc['instances']))
            return 1

        payload = client.restart_app(app_id, force)

        message = 'Created deployment {}'.format(payload['deploymentId'])
        emitter.publish(message)
        return 0

    def kill(self, app_id, scale, host):
        """
        :param app_id: the id of the application
        :type app_id: str
        :param scale: Scale the app down
        :type scale: bool
        :param host: Kill only those tasks running on host specified
        :type host: str
        :returns: process return code
        :rtype: int
        """
        client = self._create_marathon_client()

        payload = client.kill_tasks(app_id, host=host, scale=scale)
        # If scale is provided, the API return a "deploymentResult"
        # https://github.com/mesosphere/marathon/blob/50366c8/src/main/scala/mesosphere/marathon/api/RestResource.scala#L34-L36
        if scale:
            emitter.publish("Started deployment: {}".format(payload))
        else:
            if 'tasks' in payload:
                emitter.publish('Killed tasks: {}'.format(payload['tasks']))
                if len(payload['tasks']) == 0:
                    return 1
            else:
                emitter.publish('Killed tasks: []')
                return 1
        return 0

    def version_list(self, app_id, max_count):
        """
        :param app_id: the id of the application
        :type app_id: str
        :param max_count: the maximum number of version to fetch and return
        :type max_count: str
        :returns: process return code
        :rtype: int
        """

        if max_count is not None:
            max_count = util.parse_int(max_count)

        client = self._create_marathon_client()

        versions = client.get_app_versions(app_id, max_count)

        emitter.publish(versions)
        return 0

    def deployment_list(self, app_id, json_):
        """
        :param app_id: the application id
        :type app_id: str
        :param json_: output json if True
        :type json_: bool
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()

        deployments = client.get_deployments(app_id)

        if not deployments and not json_:
            msg = "There are no deployments"
            if app_id:
                msg += " for '{}'".format(app_id)
            raise DCOSException(msg)

        emitting.publish_table(emitter,
                               deployments,
                               tables.deployment_table,
                               json_)
        return 0

    def deployment_stop(self, deployment_id):
        """
        :param deployment_id: the application id
        :type deployment_id: str
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()
        client.stop_deployment(deployment_id)

        return 0

    def deployment_rollback(self, deployment_id):
        """
        :param deployment_id: the application id
        :type deployment_id: str
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()
        deployment = client.rollback_deployment(deployment_id)

        emitter.publish(deployment)
        return 0

    def deployment_watch(self, deployment_id, max_count, interval):
        """
        :param deployment_id: the application id
        :type deployment_id: str
        :param max_count: maximum number of polling calls
        :type max_count: str
        :param interval: wait interval in seconds between polling calls
        :type interval: str
        :returns: process return code
        :rtype: int
        """

        if max_count is not None:
            max_count = util.parse_int(max_count)

        interval = 1 if interval is None else util.parse_int(interval)

        client = self._create_marathon_client()

        count = 0
        while max_count is None or count < max_count:
            deployment = client.get_deployment(deployment_id)

            if deployment is None:
                return 0
            if util.is_windows_platform():
                os.system('cls')
            else:
                if 'TERM' in os.environ:
                    os.system('clear')
            emitter.publish('Deployment update time: '
                            '{} \n'.format(time.strftime("%Y-%m-%d %H:%M:%S",
                                                         time.gmtime())))
            emitter.publish(deployment)
            time.sleep(interval)
            count += 1

        return 0

    def task_list(self, app_id, json_):
        """
        :param app_id: the id of the application
        :type app_id: str
        :param json_: output json if True
        :type json_: bool
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()
        tasks = client.get_tasks(app_id)

        emitting.publish_table(emitter, tasks, tables.app_task_table, json_)
        return 0

    def task_stop(self, task_id, wipe):
        """Stop a Marathon task

        :param task_id: the id of the task
        :type task_id: str
        :param wipe: whether to wipe persistent data and unreserve resources
        :type wipe: bool
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()
        task = client.stop_task(task_id, wipe)

        if task is None:
            raise DCOSException("Task '{}' does not exist".format(task_id))

        emitter.publish(task)
        return 0

    def task_kill(self, task_ids, scale, wipe, json_):
        """Kill one or multiple Marathon tasks

        :param task_ids: the id of the task
        :type task_ids: [str]
        :param scale: Scale the app down after killing the specified tasks
        :type scale: bool
        :param wipe: whether remove reservations and persistent volumes.
        :type wipe: bool
        :param json_: output JSON if true
        :type json_: bool
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()
        payload = client.kill_and_scale_tasks(task_ids, scale, wipe)

        def print_deployment(deployment, json_):
            if json_:
                emitter.publish(deployment)
            else:
                emitter.publish('Created deployment: {}'.format(
                    deployment['deploymentId']))

        def print_killed_tasks(payload, json_):
            if json_:
                emitter.publish(payload)
            else:
                emitter.publish('Killed tasks: {}'.format(
                    [task['id'] for task in payload['tasks']]))

        if scale:
            print_deployment(payload, json_)
        else:
            print_killed_tasks(payload, json_)

            if len(payload['tasks']) == 0:
                raise DCOSException(
                    'Failed to kill tasks. task-ids seems to be unknown')

        return 0

    def task_show(self, task_id):
        """
        :param task_id: the task id
        :type task_id: str
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()
        task = client.get_task(task_id)

        if task is None:
            raise DCOSException("Task '{}' does not exist".format(task_id))

        emitter.publish(task)
        return 0

    def pod_add(self, pod_resource_path):
        """
        :param pod_resource_path: optional file path for the pod resource
        :type pod_resource_path: str
        :returns: process return code
        :rtype: int
        """

        marathon_client = self._create_marathon_client()
        self._ensure_pods_support(marathon_client)

        pod_json = self._resource_reader.get_resource(pod_resource_path)
        deployment = marathon_client.add_pod(pod_json)
        emitter.publish('Created deployment {}'.format(deployment))
        return 0

    def pod_remove(self, pod_id, force):
        """
        :param pod_id: the Marathon ID of the pod to remove
        :type pod_id: str
        :param force: whether to override running deployments
        :type force: bool
        :returns: process return code
        :rtype: int
        """

        marathon_client = self._create_marathon_client()
        self._ensure_pods_support(marathon_client)

        marathon_client.remove_pod(pod_id, force)
        return 0

    def pod_list(self, json_):
        """
        :param json_: output JSON if true
        :type json_: bool
        :returns: process return code
        :rtype: int
        """

        marathon_client = self._create_marathon_client()
        self._ensure_pods_support(marathon_client)

        pods = marathon_client.list_pod()
        queued_apps = marathon_client.get_queued_apps()
        _enhance_row_with_overdue_information(pods, queued_apps)
        emitting.publish_table(emitter, pods, tables.pod_table, json_)
        return 0

    def pod_show(self, pod_id):
        """Show details of a Marathon pod.

        :param pod_id: the Marathon ID of the pod to show
        :type pod_id: str
        :returns: process return code
        :rtype: int
        """

        marathon_client = self._create_marathon_client()
        self._ensure_pods_support(marathon_client)

        pod_json = marathon_client.show_pod(pod_id)
        emitter.publish(pod_json)
        return 0

    def pod_update(self, pod_id, force):
        """
        :param pod_id: the Marathon ID of the pod to update
        :type pod_id: str
        :param force: whether to override running deployments
        :type force: bool
        :returns: process return code
        :rtype: int
        """

        marathon_client = self._create_marathon_client()
        self._ensure_pods_support(marathon_client)

        # Ensure that the pod exists
        marathon_client.show_pod(pod_id)

        resource = self._resource_reader.get_resource(name=None)
        deployment_id = marathon_client.update_pod(
            pod_id, pod_json=resource, force=force)

        emitter.publish('Created deployment {}'.format(deployment_id))
        return 0

    def pod_kill(self, pod_id, instance_ids):
        """
        :param pod_id: the Marathon ID of the pod to kill instances from
        :type pod_id: str
        :param instance_ids: the instance IDs to kill
        :type instance_ids: [str]
        :returns: process return code
        :rtype: int
        """

        if not instance_ids:
            raise DCOSException('Please provide at least one pod instance ID')

        marathon_client = self._create_marathon_client()
        self._ensure_pods_support(marathon_client)

        marathon_client.kill_pod_instances(pod_id, instance_ids)
        return 0

    def debug_list(self, json_):
        """
        :param json_: output json if True
        :type json_: bool
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()
        queued_apps = client.get_queued_apps()

        emitting.publish_table(emitter, queued_apps,
                               tables.queued_apps_table, json_)
        return 0

    def debug_details(self, app_id, json_):
        """
        :param app_id: the Marathon ID to display details
        :type app_id: string
        :param json_: output json if True
        :type json_: bool
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()
        queued_app = client.get_queued_app(app_id)

        if queued_app:
            emitting.publish_table(
                emitter, queued_app,
                tables.queued_app_details_table, json_)
        else:
            raise DCOSException("No apps found in Marathon queue")

        return 0

    def debug_summary(self, app_id, json_):
        """
        :param app_id: the Marathon ID to display details
        :type app_id: string
        :param json_: output json if True
        :type json_: bool
        :returns: process return code
        :rtype: int
        """

        client = self._create_marathon_client()
        queued_app = client.get_queued_app(app_id)

        if queued_app:
            if queued_app.get('processedOffersSummary'):
                emitting.publish_table(
                    emitter, queued_app,
                    tables.queued_app_table, json_)
            else:
                msg = "This command is not supported on your cluster"
                raise DCOSException(msg)
        else:
            raise DCOSException("No apps found in Marathon queue")

        return 0

    @staticmethod
    def _ensure_pods_support(marathon_client):
        """Raises an exception if the given client is communicating with a
        version of Marathon that doesn't support pods.

        :param marathon_client: the Marathon client to check
        :type marathon_client: dcos.marathon.Client
        :rtype: None
        """

        if not marathon_client.pod_feature_supported():
            msg = 'This command is not supported by your version of Marathon'
            raise DCOSException(msg)


def _enhance_row_with_overdue_information(rows, queued_apps):
    """Calculates if configured `backoff` duration for this
    app or pod definition was exceeded. In that case this application
    is marked as `overdue`.
    If the app or pod inside this row should be

    :param rows: list of rows
    :type rows: []
    :param queued_apps: list of
    :type queued_apps: []
    :returns: true if pod is overdue, false otherwise
    :rtype: bool
    """

    for row in rows:
        queued_app = next(
            (app for app in queued_apps
             if row.get('id') == marathon.get_app_or_pod_id(app)),
            None)
        row['overdue'] = queued_app.get('delay', {}) \
            .get('overdue', False) if queued_app else False

    return rows


def _calculate_version(client, app_id, version):
    """
    :param client: Marathon client
    :type client: dcos.marathon.Client
    :param app_id: The ID of the application
    :type app_id: str
    :param version: Relative or absolute version or None
    :type version: str
    :returns: The absolute version as an ISO8601 date-time
    :rtype: str
    """

    # First let's try to parse it as a negative integer
    try:
        value = util.parse_int(version)
    except DCOSException:
        logger.exception('Unable to parse version %s', version)
        return version
    else:
        if value < 0:
            value = -1 * value
            # We have a negative value let's ask Marathon for the last
            # abs(value)
            versions = client.get_app_versions(app_id, value + 1)

            if len(versions) <= value:
                # We don't have enough versions. Return an error.
                msg = "Application {!r} only has {!r} version(s)."
                raise DCOSException(msg.format(app_id, len(versions), value))
            else:
                return versions[value]

        else:
            raise DCOSException(
                'Relative versions must be negative: {}'.format(version))


def _cli_config_schema():
    """
    :returns: schema for marathon cli config
    :rtype: dict
    """
    return json.loads(
        pkg_resources.resource_string(
            'dcos',
            'data/config-schema/marathon.json').decode('utf-8'))



import traceback

import pkg_resources


# must also add subcommand name to dcos.subcommand.default_subcommands
def _default_modules():
    """Dict of the default dcos cli subcommands and their main methods

    :returns: default subcommand -> main method
    :rtype: {}
    """

    # avoid circular imports
    from dcoscli.auth import main as auth_main
    from dcoscli.cluster import main as cluster_main
    from dcoscli.config import main as config_main
    from dcoscli.experimental import main as experimental_main
    from dcoscli.help import main as help_main
    from dcoscli.job import main as job_main
    from dcoscli.marathon import main as marathon_main
    from dcoscli.node import main as node_main
    from dcoscli.package import main as package_main
    from dcoscli.service import main as service_main
    from dcoscli.task import main as task_main

    return {'auth': auth_main,
            'cluster': cluster_main,
            'config': config_main,
            'experimental': experimental_main,
            'help': help_main,
            'job': job_main,
            'marathon': marathon_main,
            'node': node_main,
            'package': package_main,
            'service': service_main,
            'task': task_main
            }


def default_doc(command):
    """Returns documentation of command

    :param command: default DC/OS CLI command
    :type command: str
    :returns: config schema of command
    :rtype: str
    """

    resource = "data/help/{}.txt".format(command)
    return pkg_resources.resource_string(
        'dcoscli',
        resource).decode('utf-8')


def default_command_info(command):
    """top level documentation of default DC/OS CLI command

    :param command: name of command
    :param type: str
    :returns: command summary
    :rtype: str
    """

    doc = default_command_documentation(command)
    return doc.splitlines()[1].strip(".").lstrip()


def default_command_documentation(command):
    """documentation of default DC/OS CLI command

    :param command: name of command
    :param type: str
    :returns: command summary
    :rtype: str
    """

    return default_doc(command).rstrip('\r\n')


class SubcommandMain():

    def __init__(self, command, args):
        """Representes a subcommand running in the main thread

        :param commad: function to run in thread
        :type command: str
        """

        self._command = command
        self._args = args

    def run_and_capture(self):
        """
        Run a command and capture exceptions. This is a blocking call
        :returns: tuple of exitcode, error (or None)
        :rtype: int, str | None
        """

        m = _default_modules()[self._command]
        err = None
        try:
            exit_code = m.main([self._command] + self._args)
        except Exception:
            err = traceback.format_exc()
            traceback.print_exc()
            exit_code = 1
        return exit_code, err



import contextlib
import sys

import mock
import six


@contextlib.contextmanager
def mock_args(args):
    """ Context manager that mocks sys.args and captures stdout/stderr

    :param args: sys.args values to mock
    :type args: [str]
    :rtype: None
    """
    with mock.patch('sys.argv', ['dcos'] + args):
        stdout, stderr = sys.stdout, sys.stderr
        sys.stdout, sys.stderr = six.StringIO(), six.StringIO()
        try:
            yield sys.stdout, sys.stderr
        finally:
            sys.stdout, sys.stderr = stdout, stderr


def mock_called_some_args(mock, *args, **kwargs):
    """Convience method for some mock assertions.  Returns True if the
    arguments to one of the calls of `mock` contains `args` and
    `kwargs`.

    :param mock: the mock to check
    :type mock: mock.Mock
    :returns: True if the arguments to one of the calls for `mock`
    contains `args` and `kwargs`.
    :rtype: bool
    """

    for call in mock.call_args_list:
        call_args, call_kwargs = call

        if any(arg not in call_args for arg in args):
            continue

        if any(k not in call_kwargs or call_kwargs[k] != v
               for k, v in kwargs.items()):
            continue

        return True

    return False


def exec_mock(main, args):
    """Call a main function with sys.args mocked, and capture
    stdout/stderr

    :param main: main function to call
    :type main: function
    :param args: sys.args to mock, excluding the initial 'dcos'
    :type args: [str]
    :returns: (returncode, stdout, stderr)
    :rtype: (int, bytes, bytes)
    """

    print('MOCK ARGS: {}'.format(' '.join(args)))

    with mock_args(args) as (stdout, stderr):
        returncode = main(args)

    stdout_val = six.b(stdout.getvalue())
    stderr_val = six.b(stderr.getvalue())

    print('STDOUT: {}'.format(stdout_val))
    print('STDERR: {}'.format(stderr_val))

    return (returncode, stdout_val, stderr_val)


def assert_mock(main,
                args,
                returncode=0,
                stdout=b'',
                stderr=b''):
    """Mock and call a main function, and assert expected behavior.

    :param main: main function to call
    :type main: function
    :param args: sys.args to mock, excluding the initial 'dcos'
    :type args: [str]
    :type returncode: int
    :param stdout: Expected stdout
    :type stdout: str
    :param stderr: Expected stderr
    :type stderr: str
    :rtype: None
    """

    returncode_, stdout_, stderr_ = exec_mock(main, args)

    assert returncode_ == returncode
    assert stdout_ == stdout
    assert stderr_ == stderr



from concurrent.futures import ThreadPoolExecutor

import docopt

import dcoscli
from dcos import cmds, emitting, options, subcommand, subprocess, util
from dcos.errors import DCOSException
from dcoscli.subcommand import (default_command_documentation,
                                default_command_info, default_doc)
from dcoscli.util import decorate_docopt_usage

emitter = emitting.FlatEmitter()
logger = util.get_logger(__name__)


def main(argv):
    try:
        return _main(argv)
    except DCOSException as e:
        emitter.publish(e)
        return 1


@decorate_docopt_usage
def _main(argv):
    args = docopt.docopt(
        default_doc("help"),
        argv=argv,
        version='dcos-help version {}'.format(dcoscli.version))

    return cmds.execute(_cmds(), args)


def _cmds():
    """
    :returns: All of the supported commands
    :rtype: list of dcos.cmds.Command
    """

    return [
        cmds.Command(
            hierarchy=['help', '--info'],
            arg_keys=[],
            function=_info),

        cmds.Command(
            hierarchy=['help'],
            arg_keys=['<subcommand>'],
            function=_help),
    ]


def _info():
    """
    :returns: process return code
    :rtype: int
    """

    emitter.publish(default_command_info("help"))
    return 0


def _help(command):
    """
    :param command: the command name for which you want to see a help
    :type command: str
    :returns: process return code
    :rtype: int
    """

    if command is not None:
        _help_command(command)
    else:
        emitter.publish(default_command_documentation("dcos"))
        return 0


def dcos_help():
    """
    help text for `dcos` command

    :returns: process return code
    :rtype: int
    """

    logger.debug("DCOS bin path: {!r}".format(util.dcos_bin_path()))

    results = [(c, default_command_info(c))
               for c in subcommand.default_subcommands()]
    paths = subcommand.list_paths()
    with ThreadPoolExecutor(max_workers=max(len(paths), 1)) as executor:
        results += list(executor.map(subcommand.documentation, paths))
        commands_message = options\
            .make_command_summary_string(sorted(results))

    emitter.publish(
        "Command line utility for the Mesosphere Datacenter Operating\n"
        "System (DC/OS). The Mesosphere DC/OS is a distributed operating\n"
        "system built around Apache Mesos. This utility provides tools\n"
        "for easy management of a DC/OS installation.\n")
    emitter.publish("Available DC/OS commands:")
    emitter.publish(commands_message)
    emitter.publish(
        "\nGet detailed command description with 'dcos <command> --help'.")

    return 0


def _help_command(command):
    """
    :param command: the command name for which you want to see a help
    :type command: str
    :returns: process return code
    :rtype: int
    """

    if command in subcommand.default_subcommands():
        emitter.publish(default_command_documentation(command))
        return 0
    else:
        executable = subcommand.command_executables(command)
        return subprocess.Subproc().call([executable, command, '--help'])



import getpass
import sys
import textwrap
import time
import webbrowser

import jwt

from six.moves import urllib
from six.moves.urllib.parse import urlparse

from dcos import config, http, util
from dcos.errors import (DCOSAuthenticationException, DCOSException,
                         DCOSHTTPException)

logger = util.get_logger(__name__)


def _get_auth_scheme(response):
    """Return authentication scheme requested by server for
       'acsjwt' (DC/OS acs auth), 'oauthjwt' (DC/OS acs oauth), or
       None (no auth)

    :param response: requests.response
    :type response: requests.Response
    :returns: auth_scheme
    :rtype: str | None
    """

    if 'www-authenticate' in response.headers:
        auths = response.headers['www-authenticate'].split(',')
        scheme = next((auth_type.rstrip().lower() for auth_type in auths
                       if auth_type.rstrip().lower().startswith("acsjwt") or
                       auth_type.rstrip().lower().startswith("oauthjwt")),
                      None)
        if scheme:
            scheme_info = scheme.split("=")
            auth_scheme = scheme_info[0].split(" ")[0].lower()
            return auth_scheme
        else:
            msg = ("Server responded with an HTTP 'www-authenticate' field of "
                   "'{}', DC/OS only supports ['oauthjwt', 'acsjwt']".format(
                       response.headers['www-authenticate']))
            raise DCOSException(msg)
    else:
        logger.debug("HTTP response: no www-authenticate field found")
        return


def _prompt_user_for_token(url, token_type):
    """Get Token from user

    :param url: url for user to go to
    :type url: str
    :param token_type: type of token to be received
    :type token_type: str
    :returns: token show to user by browser
    :rtype: str
    """

    msg = textwrap.dedent("""\
        If your browser didn't open, please go to the following link:

            {url}

        Enter {token_type}: """)
    msg = msg.lstrip().format(url=url, token_type=token_type)

    try:
        webbrowser.open_new_tab(url)
    except webbrowser.Error as exc:
        logger.warning(
            'Exception occurred while calling webbrowser.open(%r): %s',
            url, exc,
        )
        pass
    sys.stderr.write(msg)
    sys.stderr.flush()
    token = sys.stdin.readline().strip()
    return token


def _get_dcostoken_by_post_with_creds(dcos_url, creds):
    """
    Get DC/OS Authentication token by POST to `acs/api/v1/auth/login`
    with specific credentials. Credentials can be uid/password for
    username/password authentication, OIDC ID token for implicit OIDC flow
    (used for open DC/OS), or uid/token for service accounts (where token is a
    jwt token encoded with private key).

    :param dcos_url: url to cluster
    :type dcos_url: str
    :param creds: credentials to login endpoint
    :type creds: {}
    :returns: DC/OS Authentication Token
    :rtype: str
    """

    url = dcos_url.rstrip('/') + '/acs/api/v1/auth/login'
    response = http._request('post', url, json=creds)

    token = None
    if response.status_code == 200:
        token = response.json()['token']
        config.set_val("core.dcos_acs_token", token)

    return token


def _prompt_for_uid_password(username, hostname):
    """Get username/password for auth

    :param username: username user for authentication
    :type username: str
    :param hostname: hostname for credentials
    :type hostname: str
    :returns: username, password
    :rtype: str, str
    """

    if username is None:
        sys.stdout.write("{}'s username: ".format(hostname))
        sys.stdout.flush()
        username = sys.stdin.readline().strip()

    password = getpass.getpass("{}@{}'s password: ".format(username, hostname))

    return username, password


def _get_dcostoken_by_dcos_uid_password_auth(
        dcos_url, username=None, password=None):
    """
    Get DC/OS Authentication Token by DC/OS uid password auth

    :param dcos_url: url to cluster
    :type dcos_url: str
    :param username: username to auth with
    :type username: str
    :param password: password to auth with
    :type password: str
    :returns: DC/OS Authentication Token if successful login
    :rtype: str
    """

    url = urlparse(dcos_url)
    hostname = url.hostname
    username = username or url.username
    password = password or url.password

    if password is None:
        username, password = _prompt_for_uid_password(username, hostname)

    creds = {"uid": username, "password": password}
    return _get_dcostoken_by_post_with_creds(dcos_url, creds)


def dcos_uid_password_auth(dcos_url, username=None, password=None):
    """
    Authenticate user using DC/OS uid password auth

    Raises exception if authentication fails.

    :param dcos_url: url to cluster
    :type dcos_url: str
    :param username: username to auth with
    :type username: str
    :param password: password to auth with
    :type password: str
    :rtype: None
    """

    dcos_token = _get_dcostoken_by_dcos_uid_password_auth(
            dcos_url, username, password)
    if not dcos_token:
        raise DCOSException("Authentication failed")
    else:
        return


def dcos_cred_auth(dcos_url, username=None, password=None):
    """
    Authenticate user using DC/OS credentials

    Raises exception if authentication fails.

    :param dcos_url: url to cluster
    :type dcos_url: str
    :param username: username to auth with
    :type username: str
    :param password: password to auth with
    :type password: str
    :rtype: None
    """

    dcos_token = _get_dcostoken_by_dcos_uid_password_auth(
            dcos_url, username, password)
    if not dcos_token:
        raise DCOSException("Authentication failed")
    else:
        return


def oidc_implicit_flow_auth(dcos_url):
    """
    Authenticate user using OIDC implict flow

    Raises exception if authentication fails.

    :param dcos_url: url to cluster
    :type dcos_url: str
    :rtype: None
    """

    dcos_auth_token = _get_dcostoken_by_oidc_implicit_flow(dcos_url)
    if not dcos_auth_token:
        raise DCOSException("Authentication failed")
    else:
        return


def _get_dcostoken_by_oidc_implicit_flow(dcos_url):
    """
    Get DC/OS Authentication Token by OIDC implicit flow

    :param dcos_url: url to cluster
    :type dcos_url: str
    :returns: DC/OS Authentication Token
    :rtype: str
    """

    oauth_login = '/login?redirect_uri=urn:ietf:wg:oauth:2.0:oob'
    url = dcos_url.rstrip('/') + oauth_login
    oidc_token = _prompt_user_for_token(url, "OpenID Connect ID Token")
    creds = {"token": oidc_token}
    return _get_dcostoken_by_post_with_creds(dcos_url, creds)


def servicecred_auth(dcos_url, username, key_path):
    """
    Get DC/OS Authentication token by browser prompt

    :param dcos_url: url to cluster
    :type dcos_url: str
    :param username: username user for authentication
    :type username: str
    :param key_path: path to service key
    :param key_path: str
    :rtype: None
    """

    # 'token' below contains a short lived service login token. This requires
    # the local machine to be in sync with DC/OS nodes enough that the 5min
    # padding here is enough time to validate the token.
    creds = {
        'uid': username,
        'token': jwt.encode(
            {
                'exp': int(time.time()+5*60),
                'uid': username
            },
            util.read_file_secure(key_path),
            algorithm='RS256')
        .decode('ascii')
    }

    dcos_token = _get_dcostoken_by_post_with_creds(dcos_url, creds)
    if not dcos_token:
        raise DCOSException("Authentication failed")
    else:
        return


def browser_prompt_auth(dcos_url, provider_info):
    """
    Get DC/OS Authentication token by browser prompt

    :param dcos_url: url to cluster
    :type dcos_url: str
    :param provider_info: info about provider to auth with
    :param provider_info: str
    :rtype: None
    """

    start_flow_url = provider_info["config"]["start_flow_url"].lstrip('/')
    if not urlparse(start_flow_url).netloc:
        start_flow_url = dcos_url.rstrip('/') + start_flow_url

    dcos_token = _prompt_user_for_token(
        start_flow_url, "DC/OS Authentication Token")

    # verify token
    endpoint = '/pkgpanda/active.buildinfo.full.json'
    url = urllib.parse.urljoin(dcos_url, endpoint)
    response = http._request('HEAD', url, auth=http.DCOSAcsAuth(dcos_token))
    if response.status_code in [200, 403]:
        config.set_val("core.dcos_acs_token", dcos_token)
    else:
        raise DCOSException("Authentication failed")


def header_challenge_auth(dcos_url):
    """
    Triggers authentication using scheme specified in www-authenticate header.

    Raises exception if authentication fails.

    :param dcos_url: url to cluster
    :type dcos_url: str
    :rtype: None
    """

    # hit protected endpoint which will prompt for auth if cluster has auth
    endpoint = '/pkgpanda/active.buildinfo.full.json'
    url = urllib.parse.urljoin(dcos_url, endpoint)
    response = http._request('HEAD', url)
    auth_scheme = _get_auth_scheme(response)

    for _ in range(3):
        if response.status_code == 401:
            # this header claims the cluster is open DC/OS 1.7, 1.8 or 1.9
            # and supports OIDC implicit auth
            if auth_scheme == "oauthjwt":
                token = _get_dcostoken_by_oidc_implicit_flow(dcos_url)
            # auth_scheme == "acsjwt"
            # this header claims the cluster is enterprise DC/OS 1.7, 1.8 or
            # 1.9 and supports username/pawword auth
            else:
                token = _get_dcostoken_by_dcos_uid_password_auth(dcos_url)

            if token is not None:
                break
        elif response.status_code == 200:
            break
    else:
        raise DCOSAuthenticationException(response)


def get_providers():
    """
    Returns dict of providers configured on cluster

    :returns: configured providers
    :rtype: {}
    """

    dcos_url = config.get_config_val("core.dcos_url").rstrip('/')
    endpoint = '/acs/api/v1/auth/providers'
    url = urllib.parse.urljoin(dcos_url, endpoint)
    try:
        providers = http.get(url)
        return providers.json()
    # this endpoint should only have authentication in DC/OS 1.8
    except DCOSAuthenticationException:
        msg = "This command is not supported for your cluster"
        raise DCOSException(msg)
    except DCOSHTTPException as e:
        if e.response.status_code == 404:
            msg = "This command is not supported for your cluster"
            raise DCOSException(msg)

    return {}


def auth_type_description(provider_info):
    """
    Returns human readable description of auth type

    :param provider_info: info about auth provider
    :type provider_info: dict
    :returns: human readable description of auth type
    :rtype: str
    """

    auth_type = provider_info.get("authentication-type")
    if auth_type == "dcos-uid-password":
        msg = ("Authenticate using a standard DC/OS user account "
               "(using username and password)")
    elif auth_type == "dcos-uid-servicekey":
        msg = ("Authenticate using a DC/OS service user account "
               "(using username and private key)")
    elif auth_type == "dcos-uid-password-ldap":
        msg = ("Authenticate using an LDAP user account "
               "(using username and password)")
    elif auth_type == "saml-sp-initiated":
        msg = "Authenticate using SAML 2.0 ({})".format(
                provider_info["description"])
    elif auth_type in ["oidc-authorization-code-flow", "oidc-implicit-flow"]:
        msg = "Authenticate using OpenID Connect ({})".format(
                provider_info["description"])
    else:
        raise DCOSException("Unknown authentication type")

    return msg



import json

from .helpers.common import assert_command, exec_command


def test_info():
    stdout = b'Manage your DC/OS clusters\n'
    assert_command(['dcos', 'cluster', '--info'],
                   stdout=stdout)


def test_version():
    stdout = b'dcos-cluster version SNAPSHOT\n'
    assert_command(['dcos', 'cluster', '--version'],
                   stdout=stdout)


def test_list():
    returncode, stdout, stderr = exec_command(
        ['dcos', 'cluster', 'list', '--json'])
    assert returncode == 0
    assert stderr == b''
    cluster_list = json.loads(stdout.decode('utf-8'))
    assert len(cluster_list) == 1
    info = cluster_list[0]
    assert info.get("attached")
    keys = ["attached", "cluster_id", "name", "url", "version"]
    assert sorted(info.keys()) == keys


def test_rename():
    _, stdout, _ = exec_command(
        ['dcos', 'cluster', 'list', '--json'])
    info = json.loads(stdout.decode('utf-8'))[0]
    name = info.get("name")

    new_name = "test"
    assert_command(['dcos', 'cluster', 'rename', name, new_name])

    returncode, stdout, stderr = exec_command(
        ['dcos', 'cluster', 'list', '--json'])
    assert json.loads(stdout.decode('utf-8'))[0].get("name") == new_name

    # rename back to original name
    assert_command(['dcos', 'cluster', 'rename', new_name, name])



from __future__ import print_function

import hashlib
import json
import os
import platform
import shutil
import stat
import subprocess
import sys
import zipfile
from distutils.version import LooseVersion

import requests

from dcos import config, constants, util
from dcos.errors import DCOSException
from dcos.subprocess import Subproc

logger = util.get_logger(__name__)


def command_executables(subcommand):
    """List the real path to executable dcos program for specified subcommand.

    :param subcommand: name of subcommand. E.g. marathon
    :type subcommand: str
    :returns: the dcos program path
    :rtype: str
    """

    executables = []
    if subcommand in default_subcommands():
        executables += [default_list_paths()]

    executables += [
        command_path
        for command_path in list_paths()
        if noun(command_path) == subcommand
    ]

    if len(executables) > 1:
        msg = 'Found more than one executable for command {!r}. {!r}'
        raise DCOSException(msg.format(subcommand, executables))

    if len(executables) == 0:
        msg = "{!r} is not a dcos command."
        raise DCOSException(msg.format(subcommand))

    return executables[0]


def get_package_commands(package_name):
    """List the real path(s) to executables for a specific dcos subcommand

    :param package_name: package name
    :type package_name: str
    :returns: list of all the dcos program paths in package
    :rtype: [str]
    """

    bin_dir = os.path.join(_package_dir(package_name),
                           constants.DCOS_SUBCOMMAND_ENV_SUBDIR,
                           BIN_DIRECTORY)

    executables = []
    for filename in os.listdir(bin_dir):
        path = os.path.join(bin_dir, filename)

        if (filename.startswith(constants.DCOS_COMMAND_PREFIX) and
                _is_executable(path)):

            executables.append(path)

    return executables


def default_list_paths():
    """List the real path to dcos executable

    :returns: list dcos program path
    :rtype: str
    """

    # Let's get all the default subcommands
    binpath = util.dcos_bin_path()
    return os.path.join(binpath, "dcos")


def list_paths():
    """List the real path to executable dcos subcommand programs.

    :returns: list of all the dcos program paths
    :rtype: [str]
    """

    subcommands = []
    for package in distributions():
        subcommands += get_package_commands(package)

    return subcommands


def _is_executable(path):
    """
    :param path: the path to a program
    :type path: str
    :returns: True if the path is an executable; False otherwise
    :rtype: bool
    """

    return os.access(path, os.X_OK) and (
        not util.is_windows_platform() or path.endswith('.exe'))


def _find_distributions(subcommand_dir):
    """
    :param subcommand_dir: directory to find packaged in
    :type subcommand_dir: path
    :returns: list of all installed subcommands in given directory
    :rtype: [str]
    """

    if subcommand_dir and os.path.isdir(subcommand_dir):
        return [
            subdir for subdir in os.listdir(subcommand_dir)
            if os.path.isdir(
                os.path.join(
                    subcommand_dir,
                    subdir,
                    constants.DCOS_SUBCOMMAND_ENV_SUBDIR))
        ]
    else:
        return []


def distributions():
    """Set of all of the installed subcommand packages

    :returns: a set of packages
    :rtype: Set[str]
    """

    cluster_packages = _find_distributions(_cluster_subcommand_dir())
    global_packages = _find_distributions(global_subcommand_dir())
    return set(cluster_packages + global_packages)


# must also add subcommand name to dcoscli.subcommand._default_modules
def default_subcommands():
    """List the default dcos cli subcommands

    :returns: list of all the default dcos cli subcommands
    :rtype: [str]
    """

    return ["auth", "cluster", "config", "experimental", "help", "job",
            "marathon", "node", "package", "service", "task"]


def documentation(executable_path):
    """Gather subcommand summary

    :param executable_path: real path to the dcos subcommands
    :type executable_path: str
    :returns: subcommand and its summary
    :rtype: (str, str)
    """

    path_noun = noun(executable_path)
    return (path_noun, info(executable_path, path_noun))


def info(executable_path, path_noun):
    """Collects subcommand information

    :param executable_path: real path to the dcos subcommand
    :type executable_path: str
    :param path_noun: subcommand
    :type path_noun: str
    :returns: the subcommand information
    :rtype: str
    """

    out = Subproc().check_output(
        [executable_path, path_noun, '--info'])

    return out.decode('utf-8').strip()


def config_schema(executable_path, noun=None):
    """Collects subcommand config schema

    :param executable_path: real path to the dcos subcommand
    :type executable_path: str
    :param noun: name of subcommand
    :type noun: str
    :returns: the subcommand config schema
    :rtype: dict
    """
    if noun is None:
        noun = noun(executable_path)

    out = Subproc().check_output(
        [executable_path, noun, '--config-schema'])

    return json.loads(out.decode('utf-8'))


def noun(executable_path):
    """Extracts the subcommand single noun from the path to the executable.
    E.g for :code:`bin/dcos-subcommand` this method returns :code:`subcommand`.

    :param executable_path: real pth to the dcos subcommand
    :type executable_path: str
    :returns: the subcommand
    :rtype: str
    """

    basename = os.path.basename(executable_path)
    noun = basename[len(constants.DCOS_COMMAND_PREFIX):].replace('.exe', '')
    return noun


def _write_package_json(pkg, pkg_dir):
    """ Write package.json locally.

    :param pkg: the package being installed
    :type pkg: PackageVersion
    :param pkg_dir: directory to install package
    :type pkg_dir: str
    :rtype: None
    """

    package_path = os.path.join(pkg_dir, 'package.json')

    package_json = pkg.package_json()

    with util.open_file(package_path, 'w') as package_file:
        json.dump(package_json, package_file)


def _hashfile(filename):
    """Calculates the sha256 of a file

    :param filename: path to the file to sum
    :type filename: str
    :returns: digest in hexadecimal
    :rtype: str
    """

    hasher = hashlib.sha256()
    with open(filename, 'rb') as f:
        for chunk in iter(lambda: f.read(4096), b''):
            hasher.update(chunk)
    return hasher.hexdigest()


def _check_hash(filename, content_hashes):
    """Validates whether downloaded binary matches expected hash

    :param filename: path to binary
    :type filename: str
    :param content_hashes: list of hash algorithms/value
    :type content_hashes: [{"algo": <str>, "value": <str>}]
    :returns: None if valid hash, else throws exception
    :rtype: None
    """

    content_hash = next((contents for contents in content_hashes
                        if contents.get("algo") == "sha256"),
                        None)
    if content_hash:
        expected_value = content_hash.get("value")
        actual_value = _hashfile(filename)
        if expected_value != actual_value:
            raise DCOSException(
                "The hash for the downloaded subcommand [{}] "
                "does not match the expected value [{}]. Aborting...".format(
                    actual_value, expected_value))
        else:
            return
    else:
        raise DCOSException(
            "Hash algorithm specified is unsupported. "
            "Please contact the package maintainer. Aborting...")


def _get_cli_binary_info(cli_resources):
    """Find compatible cli binary, if one exists

    :param cli_resources: cli property of resource.json
    :type resources: {}
    :returns: {"url": <str>, "kind": <str>, "contentHash": [{}]}
    :rtype: {} | None
    """

    if "binaries" in cli_resources:
        binaries = cli_resources["binaries"]
        arch = platform.architecture()[0]
        if arch != "64bit":
            raise DCOSException(
                "There is no compatible subcommand for your architecture [{}] "
                "We only support x86-64. Aborting...".format(arch))
        system = platform.system().lower()
        binary = binaries.get(system)
        if binary is None:
            raise DCOSException(
                "There is not compatible subcommand for your system [{}] "
                "Aborting...".format(system))
        elif "x86-64" in binary:
            return binary["x86-64"]

    raise DCOSException(
        "The CLI subcommand has unexpected format [{}]. "
        "Please contact the package maintainer. Aborting...".format(
            cli_resources))


def _install_cli(pkg, pkg_dir):
    """Install subcommand cli

    :param pkg: the package to install
    :type pkg: PackageVersion
    :param pkg_dir: directory to install package
    :type pkg_dir: str
    :rtype: None
    """

    with util.remove_path_on_error(pkg_dir) as pkg_dir:
        env_dir = os.path.join(pkg_dir, constants.DCOS_SUBCOMMAND_ENV_SUBDIR)

        resources = pkg.resource_json()

        if resources and resources.get("cli") is not None:
            binary = resources["cli"]
            binary_cli = _get_cli_binary_info(binary)
            _install_with_binary(
                pkg.name(),
                env_dir,
                binary_cli)
        elif pkg.command_json() is not None:
            install_operation = pkg.command_json()
            if 'pip' in install_operation:
                _install_with_pip(
                    pkg.name(),
                    env_dir,
                    install_operation['pip'])
            else:
                raise DCOSException(
                    "Installation methods '{}' not supported".format(
                        install_operation.keys()))
        else:
            raise DCOSException(
                "Could not find a CLI subcommand for your platform")


def install(pkg, global_=False):
    """Installs the dcos cli subcommand

    :param pkg: the package to install
    :type pkg: Package
    :param global_: whether to install the CLI globally
    :type global_: bool
    :rtype: None
    """

    if global_ or config.uses_deprecated_config():
        pkg_dir = global_package_dir(pkg.name())
    else:
        pkg_dir = _cluster_package_dir(pkg.name())

    util.ensure_dir_exists(pkg_dir)

    _write_package_json(pkg, pkg_dir)

    _install_cli(pkg, pkg_dir)


def global_subcommand_dir():
    """ Returns global subcommand dir. defaults to ~/.dcos/subcommands """

    return os.path.join(config.get_config_dir_path(),
                        constants.DCOS_SUBCOMMAND_SUBDIR)


def _cluster_subcommand_dir():
    """
    :returns: cluster specific subcommand dir or None
    :rtype: str | None
    """

    attached_cluster = config.get_attached_cluster_path()
    if attached_cluster is not None:
        return os.path.join(
            attached_cluster, constants.DCOS_SUBCOMMAND_SUBDIR)
    else:
        return None


def _cluster_package_dir(name):
    """Returns path to package directory for attached cluster

    :param name: package name
    :type name: str
    :returns: path to package directory
    :rtype: str | None
    """

    subcommand_dir = _cluster_subcommand_dir()
    if subcommand_dir is not None:
        return os.path.join(subcommand_dir, name)
    else:
        return None


def global_package_dir(name):
    """Returns path to package directory in global config

    :param name: package name
    :type name: str
    :rtype: str
    """

    return os.path.join(global_subcommand_dir(), name)


def _package_dir(name):
    """Returns cluster subcommand dir for name if exists, and if not
    returns path to global subcommand dir.

    :param name: package name
    :type name: str
    :rtype: str
    """

    cluster_subcommand = _cluster_package_dir(name)
    if cluster_subcommand and os.path.exists(cluster_subcommand):
        return cluster_subcommand
    else:
        return global_package_dir(name)


def uninstall(package_name):
    """Uninstall the dcos cli subcommand

    :param package_name: the name of the package
    :type package_name: str
    :returns: True if the subcommand was uninstalled
    :rtype: bool
    """

    pkg_dir = _package_dir(package_name)

    if os.path.isdir(pkg_dir):
        shutil.rmtree(pkg_dir)
        return True

    return False


BIN_DIRECTORY = 'Scripts' if util.is_windows_platform() else 'bin'


def _find_virtualenv(bin_directory):
    """
    :param bin_directory: directory to first use to find virtualenv
    :type bin_directory: str
    :returns: Absolute path to virutalenv program
    :rtype: str
    """

    virtualenv_path = os.path.join(bin_directory, 'virtualenv')
    if not os.path.exists(virtualenv_path):
        virtualenv_path = util.which('virtualenv')

    if virtualenv_path is None:
        msg = ("Unable to install CLI subcommand. "
               "Missing required program 'virtualenv'.\n"
               "Please see installation instructions: "
               "https://virtualenv.pypa.io/en/latest/installation.html")
        raise DCOSException(msg)

    return virtualenv_path


def _download_and_store(url, location):
    """Download given url and store in location on disk

    :param url: url to download
    :type url: str
    :param location: path to file to store url
    :type location: str
    :rtype: None
    """

    with open(location, 'wb') as f:
        r = requests.get(url, stream=True)
        for chunk in r.iter_content(1024):
            f.write(chunk)


def _install_with_binary(
        package_name,
        env_directory,
        binary_cli):
    """
    :param package_name: the name of the package
    :type package_name: str
    :param env_directory: the path to the directory in which to install the
                          package's binary_cli
    :type env_directory: str
    :param binary_cli: binary cli to install
    :type binary_cli: str
    :rtype: None
    """

    binary_url, kind = binary_cli.get("url"), binary_cli.get("kind")

    try:
        env_bin_dir = os.path.join(env_directory, BIN_DIRECTORY)

        if kind in ["executable", "zip"]:
            with util.temptext() as file_tmp:
                _, binary_tmp = file_tmp
                _download_and_store(binary_url, binary_tmp)
                _check_hash(binary_tmp, binary_cli.get("contentHash"))

                if kind == "executable":
                    util.ensure_dir_exists(env_bin_dir)
                    binary_name = "dcos-{}".format(package_name)
                    if util.is_windows_platform():
                        binary_name += '.exe'
                    binary_file = os.path.join(env_bin_dir, binary_name)

                    # copy to avoid windows error of moving open file
                    # binary_tmp will be removed by context manager
                    shutil.copy(binary_tmp, binary_file)
                else:
                    # kind == "zip"
                    with zipfile.ZipFile(binary_tmp) as zf:
                        zf.extractall(env_directory)

            # check contents for package_name/env/bin folder structure
            if not os.path.exists(env_bin_dir):
                msg = (
                    "CLI subcommand for [{}] has an unexpected format. "
                    "Please contact the package maintainer".format(
                        package_name))
                raise DCOSException(msg)
        else:
            msg = ("CLI subcommand for [{}] is an unsupported type: {}"
                   "Please contact the package maintainer".format(
                       package_name, kind))
            raise DCOSException(msg)

        # make binar(ies) executable
        for f in os.listdir(env_bin_dir):
            binary = os.path.join(env_bin_dir, f)
            if (f.startswith(constants.DCOS_COMMAND_PREFIX)):
                st = os.stat(binary)
                os.chmod(binary, st.st_mode | stat.S_IEXEC)
    except DCOSException:
        raise
    except Exception as e:
        logger.exception(e)
        raise _generic_error(package_name)

    return None


def _install_with_pip(
        package_name,
        env_directory,
        requirements):
    """
    :param package_name: the name of the package
    :type package_name: str
    :param env_directory: the path to the directory in which to install the
                          package's virtual env
    :type env_directory: str
    :param requirements: the list of pip requirements
    :type requirements: list of str
    :rtype: None
    """

    bin_directory = util.dcos_bin_path()
    new_package_dir = not os.path.exists(env_directory)

    pip_path = os.path.join(env_directory, BIN_DIRECTORY, 'pip')
    if not os.path.exists(pip_path):
        virtualenv_path = _find_virtualenv(bin_directory)

        virtualenv_version = _execute_command(
            [virtualenv_path, '--version'])[0].strip().decode('utf-8')
        if LooseVersion("12") > LooseVersion(virtualenv_version):
            msg = ("Unable to install CLI subcommand. "
                   "Required program 'virtualenv' must be version 12+, "
                   "currently version {}\n"
                   "Please see installation instructions: "
                   "https://virtualenv.pypa.io/en/latest/installation.html"
                   "".format(virtualenv_version))
            raise DCOSException(msg)

        cmd = [_find_virtualenv(bin_directory), env_directory]

        if _execute_command(cmd)[2] != 0:
            raise _generic_error(package_name)

    # Do not replace util.temptext NamedTemporaryFile
    # otherwise bad things will happen on Windows
    with util.temptext() as text_file:
        fd, requirement_path = text_file

        # Write the requirements to the file
        with os.fdopen(fd, 'w') as requirements_file:
            for line in requirements:
                print(line, file=requirements_file)

        cmd = [
            os.path.join(env_directory, BIN_DIRECTORY, 'pip'),
            'install',
            '--requirement',
            requirement_path,
        ]

        if _execute_command(cmd)[2] != 0:
            # We should remove the directory that we just created
            if new_package_dir:
                shutil.rmtree(env_directory)

            raise _generic_error(package_name)
    return None


def _execute_command(command):
    """
    :param command: a command to execute
    :type command: list of str
    :returns: stdout, stderr, the process return code
    :rtype: str, str, int
    """

    logger.info('Calling: %r', command)

    process = Subproc().popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE)

    stdout, stderr = process.communicate()

    if process.returncode != 0:
        logger.error("Command script's stdout: %s", stdout)
        logger.error("Command script's stderr: %s", stderr)
    else:
        logger.info("Command script's stdout: %s", stdout)
        logger.info("Command script's stderr: %s", stderr)

    return stdout, stderr, process.returncode


def _generic_error(package_name):
    """
    :param package: package name
    :type: str
    :returns: generic error when installing package
    :rtype: DCOSException
    """

    return DCOSException(
        ('Error installing {!r} package.\n'
         'Run with `dcos --log-level=ERROR` to see the full output.').format(
            package_name))


class InstalledSubcommand(object):
    """ Represents an installed subcommand.

    :param name: The name of the subcommand
    :type name: str
    """

    def __init__(self, name):
        self.name = name

    def _dir(self):
        """
        :returns: path to this subcommand's directory.
        :rtype: str
        """

        return _package_dir(self.name)

    def package_json(self):
        """
        :returns: contents of this subcommand's package.json file.
        :rtype: dict
        """

        package_json_path = os.path.join(self._dir(), 'package.json')
        with util.open_file(package_json_path) as package_json_file:
            return util.load_json(package_json_file)


class SubcommandProcess():

    def __init__(self, executable, command, args):
        """Representes a subcommand running by a forked process

        :param executable: executable to run
        :type executable: executable
        :param command: command to run by executable
        :type command: str
        :param args: arguments for command
        :type args: [str]
        """

        self._executable = executable
        self._command = command
        self._args = args

    def run_and_capture(self):
        """
        Run a command and capture exceptions. This is a blocking call
        :returns: tuple of exitcode, error (or None)
        :rtype: int, str | None
        """

        subproc = Subproc().popen(
            [self._executable,  self._command] + self._args,
            stderr=subprocess.PIPE)

        err = ''
        while subproc.poll() is None:
            line = subproc.stderr.readline().decode('utf-8')
            err += line
            sys.stderr.write(line)
            sys.stderr.flush()

        exitcode = subproc.poll()
        # We only want to catch exceptions, not other stderr messages
        # (such as "task does not exist", so we look for the 'Traceback'
        # string.  This only works for python, so we'll need to revisit
        # this in the future when we support subcommands written in other
        # languages.
        err = ('Traceback' in err and err) or None

        return exitcode, err



import os

import pytest

from dcos import constants

from .helpers.common import exec_command, update_config


@pytest.fixture
def env():
    r = os.environ.copy()
    r.update({
        constants.PATH_ENV: os.environ[constants.PATH_ENV],
        'DCOS_SNAKEOIL_CRT_PATH': os.environ.get(
            "DCOS_SNAKEOIL_CRT_PATH", "/dcos-cli/adminrouter/snakeoil.crt"),
        'DCOS_URL': 'https://dcos.snakeoil.mesosphere.com'
    })

    return r


def test_dont_verify_ssl_with_env_var(env):
    env['DCOS_SSL_VERIFY'] = 'false'

    returncode, stdout, stderr = exec_command(
        ['dcos', 'marathon', 'app', 'list'], env)
    assert returncode == 0
    assert stderr == b''

    env.pop('DCOS_SSL_VERIFY')


def test_dont_verify_ssl_with_config(env):
    with update_config('core.ssl_verify', 'false', env):
        returncode, stdout, stderr = exec_command(
            ['dcos', 'marathon', 'app', 'list'], env)
        assert returncode == 0
        assert stderr == b''


def test_verify_ssl_without_cert_env_var(env):
    env['DCOS_SSL_VERIFY'] = 'true'
    with update_config('core.ssl_verify', None, env):
        returncode, stdout, stderr = exec_command(
            ['dcos', 'marathon', 'app', 'list'], env)
        assert returncode == 1
        assert stderr.decode('utf-8') == _ssl_error_msg()

    env.pop('DCOS_SSL_VERIFY')


def test_verify_ssl_without_cert_config(env):
    with update_config('core.ssl_verify', 'true', env):
        returncode, stdout, stderr = exec_command(
            ['dcos', 'marathon', 'app', 'list'], env)
        assert returncode == 1
        assert stderr.decode('utf-8') == _ssl_error_msg()


def test_verify_ssl_with_bad_cert_env_var(env):
    env['DCOS_SSL_VERIFY'] = 'tests/data/ssl/fake.pem'

    with update_config('core.ssl_verify', None, env):
        returncode, stdout, stderr = exec_command(
            ['dcos', 'marathon', 'app', 'list'], env)
        assert returncode == 1
        assert stderr.decode('utf-8') == _ssl_error_msg()

    env.pop('DCOS_SSL_VERIFY')


def test_verify_ssl_with_bad_cert_config(env):
    with update_config('core.ssl_verify', 'tests/data/ssl/fake.pem', env):
        returncode, stdout, stderr = exec_command(
            ['dcos', 'marathon', 'app', 'list'], env)
        assert returncode == 1
        assert stderr.decode('utf-8') == _ssl_error_msg()


@pytest.mark.skipif(True, reason='Need to resolve DCOS-9273 to validate certs')
def test_verify_ssl_with_good_cert_env_var(env):
    env['DCOS_SSL_VERIFY'] = env['DCOS_SNAKEOIL_CRT_PATH']

    with update_config('core.ssl_verify', None, env):
        returncode, stdout, stderr = exec_command(
            ['dcos', 'marathon', 'app', 'list'], env)
        assert returncode == 0
        assert stderr == b''

    env.pop('DCOS_SSL_VERIFY')


@pytest.mark.skipif(True, reason='Need to resolve DCOS-9273 to validate certs')
def test_verify_ssl_with_good_cert_config(env):
    with update_config(
            'core.ssl_verify', env['DCOS_SNAKEOIL_CRT_PATH'], env):
        returncode, stdout, stderr = exec_command(
            ['dcos', 'marathon', 'app', 'list'], env)
        assert returncode == 0
        assert stderr == b''


def _ssl_error_msg():
    return (
        "An SSL error occurred. To configure your SSL settings, please run: "
        "`dcos config set core.ssl_verify <value>`\n"
        "<value>: Whether to verify SSL certs for HTTPS or path to certs. "
        "Valid values are True, False, or a path to a CA_BUNDLE.\n")
